@@ -42,16 +42,28 @@ class DemystConnectorError(Exception):
 
 class Connectors(object):
 
-    def __init__(self, config, inputs={}, sample_mode=False):
+    def __init__(self, config, inputs=None, sample_mode=False, master_transaction_id=None):
+        if inputs is None:
+            inputs = {}
         self.C = config
         self.key = config.get('API_KEY')
         self.inputs = inputs
         self.sample_mode = sample_mode
         self.providers_cache = {}
         self.providers_errors = {}
+        self.master_transaction_id = master_transaction_id
+
+    def provider_to_string(self, provider):
+        if isinstance(provider, str):
+            return provider
+        else:
+            return provider["name"]
 
     # XXX Hash by Inputs+BKF Options+Maybe Sample Mode
-    def cache_key(self, provider_name, inputs=None):
+    def cache_key(self, provider, inputs=None):
+        # Support versioned providers.
+        # https://github.com/DemystData/demyst-python/issues/568
+        provider_name = self.provider_to_string(provider)
         inputs = self.inputs if inputs == None else inputs
         inputs_hash = hash(frozenset(inputs.items()))
         key = provider_name + '.' + str(inputs_hash)
@@ -69,7 +81,7 @@ class Connectors(object):
             self.providers_cache[key] = provider_payload
             return True
 
-    def fetch(self, provider_names, inputs=None, sample_mode=False):
+    def fetch(self, provider_names, inputs=None, sample_mode=False, config={}):
         inputs = self.inputs if inputs == None else inputs
 
         api_key = None
@@ -80,14 +92,16 @@ class Connectors(object):
         else:
             api_key = self.C.prompt_for_jwt_token_and_cache_it()
 
-        # XXX be able to set BKF Options
         params = {
             "providers": provider_names,
             "api_key": api_key,
             "inputs": inputs,
-            "config": {}
+            "config": config
         }
 
+        if self.master_transaction_id:
+            params['master_transaction_id'] = self.master_transaction_id
+
         if (self.sample_mode or sample_mode):
             params['config']['mode'] = 'sample'
 
@@ -119,13 +133,39 @@ class Connectors(object):
         else:
             # Connector failures
             no_errors = True
+            input_errors = jresp.get('input_errors')
             for provider in jresp.get('output').keys():
                 payload = jresp.get('output').get(provider)
+                # Add input errors to error message
+                if input_errors:
+                    provider_input_errors = input_errors.get(provider)
+                    if provider_input_errors:
+                        # Normally, payload should have error object, but create one if not
+                        error = payload.get('error')
+                        if not error:
+                            error = {}
+                            payload['error'] = error
+                        msg = error.get('message') or ''
+                        msg += self.format_input_errors(provider_input_errors)
+                        error['message'] = msg
                 r = self.cache_add(provider, payload, inputs)
                 if not r:
                     no_errors = False
             return no_errors
 
+    def format_input_errors(self, errors):
+        result = ''
+        for col, col_error in errors.items():
+            result += ('\n* ' + col + ': ' + col_error['error']['message'])
+        return result
+
+    def clear_cache(self, provider_name, inputs=None):
+        inputs = self.inputs if inputs == None else inputs
+        key = self.cache_key(provider_name, inputs)
+
+        if key in self.providers_cache:
+            del self.providers_cache[key]
+
     def cache_get_error(self, provider_name, inputs=None):
         inputs = self.inputs if inputs == None else inputs
         key = self.cache_key(provider_name, inputs)
@@ -155,17 +195,17 @@ class Connectors(object):
     def raw(self, provider_name, inputs=None):
         return self.cache_get(provider_name, inputs=inputs, shape="raw")
 
-    def get(self, provider_name, query, default=None, inputs=None, shape="nested", prefix=False):
+    def get(self, provider, query, default=None, inputs=None, shape="nested", prefix=False):
         # XXX if Check fails log the miss on the cache
         # er = glom.Check(dict, query)
         # df.log(uuid, warn, er)
         if os.getenv("DEMYST_DF_PROVIDER_ERROR"):
             provider_data = DEFAULT_PRODUCT_ERROR
-        elif provider_name == 'pass_through':
+        elif provider == 'pass_through':
             # Ignore inputs parameter if pass_through is selected
             provider_data = self.inputs
         else:
-            provider_data = self.cache_get(provider_name, inputs=inputs, shape=shape)
+            provider_data = self.cache_get(provider, inputs=inputs, shape=shape)
         if query == '':
             query = T
         elif shape == "table":
@@ -175,7 +215,7 @@ class Connectors(object):
         if prefix and (shape == "table") and (type(res) is dict):
             keys = list(res).copy()
             for key in keys:
-                res[provider_name+'.'+key] = res.pop(key)
+                res[self.provider_to_string(provider)+'.'+key] = res.pop(key)
 
         return res
 
@@ -191,11 +231,17 @@ class Connectors(object):
                 del copy["version"]
                 results.append(copy)
             return results
-            
+
     def product_catalog(self, provider_name):
         p = self.C.lookup_provider(provider_name)
+        if p:
+            return self.product_catalog_of_provider(p)
+        else:
+            return None
+
+    def product_catalog_of_provider(self, p):
         if "version" in p:
             version = p["version"]
-            if "data" in version:
-                return version["data"]
+            if version and ("schema" in version):
+                return version["schema"]
         return None
